package cz.jzitnik.chronos.controllers;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import cz.jzitnik.chronos.entities.Item;
import cz.jzitnik.chronos.entities.Message;
import cz.jzitnik.chronos.entities.MessageType;
import cz.jzitnik.chronos.entities.Player;
import cz.jzitnik.chronos.payload.errors.FullInventoryError;
import cz.jzitnik.chronos.payload.errors.ItemNotUsableException;
import cz.jzitnik.chronos.payload.errors.NotFoundError;
import cz.jzitnik.chronos.payload.requests.PlayerNameRequest;
import cz.jzitnik.chronos.payload.responses.InventoryFullResponse;
import cz.jzitnik.chronos.payload.responses.ItemMoveMessageResponse;
import cz.jzitnik.chronos.payload.responses.UnifiedResponse;
import cz.jzitnik.chronos.repository.GameRepository;
import cz.jzitnik.chronos.repository.ItemRepository;
import cz.jzitnik.chronos.repository.PlayerRepository;
import cz.jzitnik.chronos.services.GameService;
import cz.jzitnik.chronos.utils.annotations.CheckUser;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

import java.util.List;

@CrossOrigin(origins = "*", maxAge = 3600)
@RestController
@RequestMapping("/game/players")
public class PlayerController {
    @Autowired
    private GameRepository gameRepository;

    @Autowired
    private PlayerRepository playerRepository;

    @Autowired
    private GameService gameService;

    @Autowired
    private ItemRepository itemRepository;

    @PostMapping("/new")
    public ResponseEntity<UnifiedResponse<String, Error>> newPlayer(@RequestParam String gameKey, @RequestBody PlayerNameRequest playerNameRequest) {
        var gameOptional = gameRepository.findByGameKey(gameKey);

        if (gameOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(
                UnifiedResponse.failure(new NotFoundError("Game with key " + gameKey + " was not found!"))
            );
        }

        var game = gameOptional.get();

        var player = gameService.addPlayer(game, playerNameRequest.getName());

        return ResponseEntity.ok(
                UnifiedResponse.success(player.getPlayerKey())
        );
    }

    @GetMapping("/me")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Player, Error>> getMe(@RequestParam String playerKey) {
        var player = playerRepository.findByPlayerKey(playerKey).get();

        return ResponseEntity.ok(UnifiedResponse.success(player));
    }

    @GetMapping("/me/inv_full")
    @CheckUser
    public ResponseEntity<UnifiedResponse<InventoryFullResponse, Error>> inventoryIsFull(@RequestParam String playerKey) {
        var player = playerRepository.findByPlayerKey(playerKey).get();

        return ResponseEntity.ok(UnifiedResponse.success(new InventoryFullResponse(player.getInventory().size() >= 3)));
    }

    @PostMapping("/use_item")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Object, Error>> useItem(@RequestParam String playerKey, @RequestParam Long itemId) {
        var player = playerRepository.findByPlayerKey(playerKey).get();
        var itemOptional = itemRepository.findById(itemId);

        if (itemOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(UnifiedResponse.failure(new NotFoundError("Item was not found!")));
        }

        var item = itemOptional.get();

        if (!player.getInventory().contains(item)) {
            return ResponseEntity.status(HttpStatus.FORBIDDEN).body(UnifiedResponse.failure(new Error("You do not own this item!")));
        }

        // Use the item
        try {
            item.getItemType().useItem(player);
        } catch (ItemNotUsableException e) {
            return ResponseEntity.status(HttpStatus.BAD_REQUEST).body(UnifiedResponse.failure(e.toError()));
        }

        player.getMessages().add(new Message(player, String.valueOf(item.getId()), MessageType.ITEM_USED));

        playerRepository.save(player);

        item.setOwner(null);

        itemRepository.save(item);

        return ResponseEntity.ok(UnifiedResponse.success(null));
    }

    @GetMapping("/item")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Item, Error>> getItemInfo(@RequestParam String playerKey, @RequestParam Long itemId) {
        var itemOptional = itemRepository.findById(itemId);

        if (itemOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(UnifiedResponse.failure(new NotFoundError("Item was not found!")));
        }
        var item = itemOptional.get();

        return ResponseEntity.ok(UnifiedResponse.success(item));
    }

    @GetMapping
    @CheckUser
    public ResponseEntity<UnifiedResponse<List<Player>, Error>> getAllPlayers(@RequestParam String playerKey) {
        var player = playerRepository.findByPlayerKey(playerKey).get();
        var game = player.getGame();

        var players = game.getPlayers();

        return ResponseEntity.ok(UnifiedResponse.success(players));
    }

    @PostMapping("/send_item")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Object, Error>> sendItem(@RequestParam String playerKey, @RequestParam Long itemId, @RequestParam Long playerId) {
        var player = playerRepository.findByPlayerKey(playerKey).get();

        var player2Optional = playerRepository.findById(playerId);
        if (player2Optional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(UnifiedResponse.failure(new NotFoundError("Player with id " + playerId + " was not found!")));
        }
        var player2 = player2Optional.get();

        var itemOptional = itemRepository.findById(itemId);
        if (itemOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(UnifiedResponse.failure(new NotFoundError("Item with id " + itemId + " was not found!")));
        }
        var item = itemOptional.get();

        if (item.getOwner() == null || !item.getOwner().getId().equals(player.getId())) {
            return ResponseEntity.status(HttpStatus.FORBIDDEN).body(UnifiedResponse.failure(new Error("You don't own this item!")));
        }
        if (player2.getInventorySize() - player2.getInventory().size() == 0) {
            // Inventory is full
            return ResponseEntity.status(HttpStatus.CONFLICT).body(UnifiedResponse.failure(new FullInventoryError("Player has full inventory!")));
        }

        item.setOwner(player2);
        itemRepository.save(item);

        ObjectMapper objectMapper = new ObjectMapper();
        try {
            var json = objectMapper.writeValueAsString(new ItemMoveMessageResponse(item.getItemType(), player2.getName()));

            var message = new Message(player, json, MessageType.ITEM_MOVED);

            player.getMessages().add(message);
            playerRepository.save(player);

            return ResponseEntity.ok(UnifiedResponse.success(null));
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }
    }

    @PostMapping("/drop_item")
    @CheckUser
    public ResponseEntity<UnifiedResponse<Object, Error>> dropItem(@RequestParam String playerKey, @RequestParam Long itemId) {
        var player = playerRepository.findByPlayerKey(playerKey).get();

        var itemOptional = itemRepository.findById(itemId);
        if (itemOptional.isEmpty()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(UnifiedResponse.failure(new NotFoundError("Item with id " + itemId + " was not found!")));
        }
        var item = itemOptional.get();
        if (!item.getOwner().getId().equals(player.getId())) {
            return ResponseEntity.status(HttpStatus.FORBIDDEN).body(UnifiedResponse.failure(new Error("You do not own this item!")));
        }

        var room = player.getCurrentRoom();
        item.setOwner(null);
        item.setRoom(room);

        itemRepository.save(item);

        return ResponseEntity.ok(UnifiedResponse.empty(true));
    }
}
