package cz.jzitnik.chronos.interactions.list;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import cz.jzitnik.chronos.entities.Character;
import cz.jzitnik.chronos.entities.Item;
import cz.jzitnik.chronos.entities.ItemType;
import cz.jzitnik.chronos.entities.Player;
import cz.jzitnik.chronos.interactions.InteractionPlayer;
import cz.jzitnik.chronos.payload.responses.InteractionResponse;
import cz.jzitnik.chronos.repository.CharacterRepository;
import cz.jzitnik.chronos.repository.ItemRepository;
import cz.jzitnik.chronos.repository.PlayerRepository;
import cz.jzitnik.chronos.services.ItemService;
import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Random;

@Service
public class NumberGuessingGame implements InteractionPlayer {
    @Autowired
    private CharacterRepository characterRepository;
    @Autowired
    private ItemService itemService;
    @Autowired
    private PlayerRepository playerRepository;
    @Autowired
    private ItemRepository itemRepository;

    private static final int RANGE_MIN = 1;
    private static final int RANGE_MAX = 100;
    private static final int MAX_ATTEMPTS = 7;

    @AllArgsConstructor
    @NoArgsConstructor
    @Getter
    @Setter
    private static class GuessingGameMemory {
        private int targetNumber;
        private int attemptsLeft;
    }

    private static GuessingGameMemory readMemory(String memory) {
        ObjectMapper objectMapper = new ObjectMapper();
        try {
            return objectMapper.readValue(memory, GuessingGameMemory.class);
        } catch (Exception e) {
            throw new RuntimeException("Error reading memory from DB.");
        }
    }

    private static String writeMemory(GuessingGameMemory memory) {
        ObjectMapper objectMapper = new ObjectMapper();
        try {
            return objectMapper.writeValueAsString(memory);
        } catch (JsonProcessingException e) {
            throw new RuntimeException("Error writing memory to DB.");
        }
    }

    @AllArgsConstructor
    @Getter
    @Setter
    private static class GuessingGameResponse {
        public enum Type {
            GAME_CREATED,
            GAME_WON,
            GAME_LOST,
            GUESS_FEEDBACK
        }

        private Type type;
        private String message;
        private int attemptsLeft;
    }

    @Override
    public InteractionResponse play(Player player, Character character, String data) {
        if (character.getInteractionData().getPlayer() == null) {
            // New game initialization
            try {
                return initGame(player, character);
            } catch (JsonProcessingException e) {
                throw new RuntimeException(e);
            }
        }

        // Ongoing game
        if (!character.getInteractionData().getPlayer().getId().equals(player.getId())) {
            return new InteractionResponse(false, "already_playing", new ArrayList<>());
        }

        try {
            return processGuess(player, character, data);
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }
    }

    private InteractionResponse initGame(Player player, Character character) throws JsonProcessingException {
        var coals = player.getInventory().stream().filter(item -> item.getItemType().equals(ItemType.COAL)).toList();
        if (coals.isEmpty()) {
            return new InteractionResponse(false, "Nemáš uhlí pro mě!", new ArrayList<>());
        }
        var coal = coals.get(0);
        coal.setOwner(null);
        itemRepository.save(coal);
        player.getInventory().remove(coal);
        playerRepository.save(player);

        var interactionData = character.getInteractionData();
        interactionData.setPlayer(player);

        Random random = new Random();
        int targetNumber = random.nextInt(RANGE_MAX - RANGE_MIN + 1) + RANGE_MIN;

        interactionData.setMemory(writeMemory(
                new GuessingGameMemory(targetNumber, MAX_ATTEMPTS)
        ));

        characterRepository.save(character);

        var response = new GuessingGameResponse(
                GuessingGameResponse.Type.GAME_CREATED,
                String.format("Uhodni číslo mezi %d a %d.", RANGE_MIN, RANGE_MAX),
                MAX_ATTEMPTS
        );

        ObjectMapper objectMapper = new ObjectMapper();
        return new InteractionResponse(true, objectMapper.writeValueAsString(response), new ArrayList<>());
    }

    private InteractionResponse processGuess(Player player, Character character, String data) throws JsonProcessingException {
        var memory = readMemory(character.getInteractionData().getMemory());

        int guess;
        try {
            guess = Integer.parseInt(data);
        } catch (NumberFormatException e) {
            return new InteractionResponse(false, "invalid_input", new ArrayList<>());
        }

        if (guess < RANGE_MIN || guess > RANGE_MAX) {
            return new InteractionResponse(false, "out_of_range", new ArrayList<>());
        }

        memory.setAttemptsLeft(memory.getAttemptsLeft() - 1);
        String feedbackMessage;
        boolean isGameWon = false;

        if (guess == memory.getTargetNumber()) {
            feedbackMessage = "Gratuluji! Uhodl jsi číslo. Tady máš fragment klíče.";
            isGameWon = true;
        } else if (guess < memory.getTargetNumber()) {
            feedbackMessage = "Tvoje číslo je moc nízké. Zkus to znovu.";
        } else {
            feedbackMessage = "Tvoje číslo je moc vysoké. Zkus to znovu.";
        }

        if (isGameWon || memory.getAttemptsLeft() <= 0) {
            character.setInteractedWith(true);
            characterRepository.save(character);

            if (isGameWon) {
                var item = new Item(ItemType.KEY_FRAGMENT, player);
                itemService.addItem(player, item);

                var items = new ArrayList<Item>();
                items.add(item);

                var response = new GuessingGameResponse(
                        GuessingGameResponse.Type.GAME_WON,
                        feedbackMessage,
                        memory.getAttemptsLeft()
                );

                ObjectMapper objectMapper = new ObjectMapper();
                return new InteractionResponse(true, objectMapper.writeValueAsString(response), items);
            } else {
                var response = new GuessingGameResponse(
                        GuessingGameResponse.Type.GAME_LOST,
                        String.format("Prohrál jsi! Správné číslo bylo %d.", memory.getTargetNumber()),
                        memory.getAttemptsLeft()
                );

                ObjectMapper objectMapper = new ObjectMapper();
                return new InteractionResponse(false, objectMapper.writeValueAsString(response), new ArrayList<>());
            }
        }

        // Update memory and save
        character.getInteractionData().setMemory(writeMemory(memory));
        characterRepository.save(character);

        var response = new GuessingGameResponse(
                GuessingGameResponse.Type.GUESS_FEEDBACK,
                feedbackMessage,
                memory.getAttemptsLeft()
        );

        ObjectMapper objectMapper = new ObjectMapper();
        return new InteractionResponse(false, objectMapper.writeValueAsString(response), new ArrayList<>());
    }
}
